import asyncio
import logging
import sys

import pytest

from prefect.settings import PREFECT_ASYNC_FETCH_STATE_RESULT, temporary_settings
from prefect.testing.utilities import prefect_test_harness


@pytest.fixture(scope="session", autouse=True)
def prefect_db():
    """
    Sets up test harness for temporary DB during test runs.
    """
    with prefect_test_harness():
        yield


@pytest.fixture(scope="session")
def event_loop(request):
    """
    Redefine the event loop to support session/module-scoped fixtures;
    see https://github.com/pytest-dev/pytest-asyncio/issues/68
    When running on Windows we need to use a non-default loop for subprocess support.
    """
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    policy = asyncio.get_event_loop_policy()

    loop = policy.new_event_loop()

    # configure asyncio logging to capture long running tasks
    asyncio_logger = logging.getLogger("asyncio")
    asyncio_logger.setLevel("WARNING")
    asyncio_logger.addHandler(logging.StreamHandler())
    loop.set_debug(True)
    loop.slow_callback_duration = 0.25

    try:
        yield loop
    finally:
        loop.close()

    # Workaround for failures in pytest_asyncio 0.17;
    # see https://github.com/pytest-dev/pytest-asyncio/issues/257
    policy.set_event_loop(loop)


@pytest.fixture(autouse=True)
def fetch_state_result():
    with temporary_settings(updates={PREFECT_ASYNC_FETCH_STATE_RESULT: True}):
        yield



"""
Interface and implementations of the Dask Task Runner.
[Task Runners](https://docs.prefect.io/api-ref/prefect/task-runners/)
in Prefect are responsible for managing the execution of Prefect task runs.
Generally speaking, users are not expected to interact with
task runners outside of configuring and initializing them for a flow.

Example:
    ```python
    import time

    from prefect import flow, task

    @task
    def shout(number):
        time.sleep(0.5)
        print(f"#{number}")

    @flow
    def count_to(highest_number):
        for number in range(highest_number):
            shout.submit(number)

    if __name__ == "__main__":
        count_to(10)

    # outputs
    #0
    #1
    #2
    #3
    #4
    #5
    #6
    #7
    #8
    #9
    ```

    Switching to a `DaskTaskRunner`:
    ```python
    import time

    from prefect import flow, task
    from prefect_dask import DaskTaskRunner

    @task
    def shout(number):
        time.sleep(0.5)
        print(f"#{number}")

    @flow(task_runner=DaskTaskRunner)
    def count_to(highest_number):
        for number in range(highest_number):
            shout.submit(number)

    if __name__ == "__main__":
        count_to(10)

    # outputs
    #3
    #7
    #2
    #6
    #4
    #0
    #1
    #5
    #8
    #9
    ```
"""

import inspect
from contextlib import ExitStack
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    Optional,
    Set,
    TypeVar,
    Union,
    overload,
)

import distributed
from typing_extensions import ParamSpec

from prefect.client.schemas.objects import State, TaskRunInput
from prefect.futures import PrefectFuture, PrefectFutureList, PrefectWrappedFuture
from prefect.logging.loggers import get_logger, get_run_logger
from prefect.task_runners import TaskRunner
from prefect.tasks import Task
from prefect.utilities.asyncutils import run_coro_as_sync
from prefect.utilities.collections import visit_collection
from prefect.utilities.importtools import from_qualified_name, to_qualified_name
from prefect_dask.client import PrefectDaskClient

logger = get_logger(__name__)

P = ParamSpec("P")
T = TypeVar("T")
F = TypeVar("F", bound=PrefectFuture)
R = TypeVar("R")


class PrefectDaskFuture(PrefectWrappedFuture[R, distributed.Future]):
    """
    A Prefect future that wraps a distributed.Future. This future is used
    when the task run is submitted to a DaskTaskRunner.
    """

    def wait(self, timeout: Optional[float] = None) -> None:
        try:
            result = self._wrapped_future.result(timeout=timeout)
        except Exception:
            # either the task failed or the timeout was reached
            return
        if isinstance(result, State):
            self._final_state = result

    def result(
        self,
        timeout: Optional[float] = None,
        raise_on_failure: bool = True,
    ) -> R:
        if not self._final_state:
            try:
                future_result = self._wrapped_future.result(timeout=timeout)
            except distributed.TimeoutError as exc:
                raise TimeoutError(
                    f"Task run {self.task_run_id} did not complete within {timeout} seconds"
                ) from exc

            if isinstance(future_result, State):
                self._final_state = future_result
            else:
                return future_result

        _result = self._final_state.result(
            raise_on_failure=raise_on_failure, fetch=True
        )
        # state.result is a `sync_compatible` function that may or may not return an awaitable
        # depending on whether the parent frame is sync or not
        if inspect.isawaitable(_result):
            _result = run_coro_as_sync(_result)
        return _result

    def __del__(self):
        if self._final_state or self._wrapped_future.done():
            return
        try:
            local_logger = get_run_logger()
        except Exception:
            local_logger = logger
        local_logger.warning(
            "A future was garbage collected before it resolved."
            " Please call `.wait()` or `.result()` on futures to ensure they resolve.",
        )


class DaskTaskRunner(TaskRunner):
    """
    A parallel task_runner that submits tasks to the `dask.distributed` scheduler.
    By default a temporary `distributed.LocalCluster` is created (and
    subsequently torn down) within the `start()` contextmanager. To use a
    different cluster class (e.g.
    [`dask_kubernetes.KubeCluster`](https://kubernetes.dask.org/)), you can
    specify `cluster_class`/`cluster_kwargs`.

    Alternatively, if you already have a dask cluster running, you can provide
    the cluster object via the `cluster` kwarg or the address of the scheduler
    via the `address` kwarg.
    !!! warning "Multiprocessing safety"
        Note that, because the `DaskTaskRunner` uses multiprocessing, calls to flows
        in scripts must be guarded with `if __name__ == "__main__":` or warnings will
        be displayed.

    Args:
        cluster (distributed.deploy.Cluster, optional): Currently running dask cluster;
            if one is not provider (or specified via `address` kwarg), a temporary
            cluster will be created in `DaskTaskRunner.start()`. Defaults to `None`.
        address (string, optional): Address of a currently running dask
            scheduler. Defaults to `None`.
        cluster_class (string or callable, optional): The cluster class to use
            when creating a temporary dask cluster. Can be either the full
            class name (e.g. `"distributed.LocalCluster"`), or the class itself.
        cluster_kwargs (dict, optional): Additional kwargs to pass to the
            `cluster_class` when creating a temporary dask cluster.
        adapt_kwargs (dict, optional): Additional kwargs to pass to `cluster.adapt`
            when creating a temporary dask cluster. Note that adaptive scaling
            is only enabled if `adapt_kwargs` are provided.
        client_kwargs (dict, optional): Additional kwargs to use when creating a
            [`dask.distributed.Client`](https://distributed.dask.org/en/latest/api.html#client).

    Examples:
        Using a temporary local dask cluster:
        ```python
        from prefect import flow
        from prefect_dask.task_runners import DaskTaskRunner

        @flow(task_runner=DaskTaskRunner)
        def my_flow():
            ...
        ```

        Using a temporary cluster running elsewhere. Any Dask cluster class should
        work, here we use [dask-cloudprovider](https://cloudprovider.dask.org):
        ```python
        DaskTaskRunner(
            cluster_class="dask_cloudprovider.FargateCluster",
            cluster_kwargs={
                "image": "prefecthq/prefect:latest",
                "n_workers": 5,
            },
        )
        ```

        Connecting to an existing dask cluster:
        ```python
        DaskTaskRunner(address="192.0.2.255:8786")
        ```
    """

    def __init__(
        self,
        cluster: Optional[distributed.deploy.Cluster] = None,
        address: Optional[str] = None,
        cluster_class: Union[str, Callable, None] = None,
        cluster_kwargs: Optional[Dict] = None,
        adapt_kwargs: Optional[Dict] = None,
        client_kwargs: Optional[Dict] = None,
    ):
        # Validate settings and infer defaults
        if address:
            if cluster or cluster_class or cluster_kwargs or adapt_kwargs:
                raise ValueError(
                    "Cannot specify `address` and "
                    "`cluster`/`cluster_class`/`cluster_kwargs`/`adapt_kwargs`"
                )
        elif cluster:
            if cluster_class or cluster_kwargs:
                raise ValueError(
                    "Cannot specify `cluster` and `cluster_class`/`cluster_kwargs`"
                )
        else:
            if isinstance(cluster_class, str):
                cluster_class = from_qualified_name(cluster_class)
            else:
                cluster_class = cluster_class

        # Create a copies of incoming kwargs since we may mutate them
        cluster_kwargs = cluster_kwargs.copy() if cluster_kwargs else {}
        adapt_kwargs = adapt_kwargs.copy() if adapt_kwargs else {}
        client_kwargs = client_kwargs.copy() if client_kwargs else {}

        # Update kwargs defaults
        client_kwargs.setdefault("set_as_default", False)

        # The user cannot specify async/sync themselves
        if "asynchronous" in client_kwargs:
            raise ValueError(
                "`client_kwargs` cannot set `asynchronous`. "
                "This option is managed by Prefect."
            )
        if "asynchronous" in cluster_kwargs:
            raise ValueError(
                "`cluster_kwargs` cannot set `asynchronous`. "
                "This option is managed by Prefect."
            )

        # Store settings
        self.address = address
        self.cluster_class = cluster_class
        self.cluster_kwargs = cluster_kwargs
        self.adapt_kwargs = adapt_kwargs
        self.client_kwargs = client_kwargs

        # Runtime attributes
        self._client: PrefectDaskClient = None
        self._cluster: "distributed.deploy.Cluster" = cluster

        self._exit_stack = ExitStack()

        super().__init__()

    def __eq__(self, other: object) -> bool:
        """
        Check if an instance has the same settings as this task runner.
        """
        if isinstance(other, DaskTaskRunner):
            return (
                self.address == other.address
                and self.cluster_class == other.cluster_class
                and self.cluster_kwargs == other.cluster_kwargs
                and self.adapt_kwargs == other.adapt_kwargs
                and self.client_kwargs == other.client_kwargs
            )
        else:
            return False

    def duplicate(self):
        """
        Create a new instance of the task runner with the same settings.
        """
        return type(self)(
            address=self.address,
            cluster_class=self.cluster_class,
            cluster_kwargs=self.cluster_kwargs,
            adapt_kwargs=self.adapt_kwargs,
            client_kwargs=self.client_kwargs,
        )

    @overload
    def submit(
        self,
        task: "Task[P, Coroutine[Any, Any, R]]",
        parameters: Dict[str, Any],
        wait_for: Optional[Iterable[PrefectFuture]] = None,
        dependencies: Optional[Dict[str, Set[TaskRunInput]]] = None,
    ) -> PrefectDaskFuture[R]:
        ...

    @overload
    def submit(
        self,
        task: "Task[Any, R]",
        parameters: Dict[str, Any],
        wait_for: Optional[Iterable[PrefectFuture]] = None,
        dependencies: Optional[Dict[str, Set[TaskRunInput]]] = None,
    ) -> PrefectDaskFuture[R]:
        ...

    def submit(
        self,
        task: Task,
        parameters: Dict[str, Any],
        wait_for: Optional[Iterable[PrefectFuture]] = None,
        dependencies: Optional[Dict[str, Set[TaskRunInput]]] = None,
    ) -> PrefectDaskFuture:
        if not self._started:
            raise RuntimeError(
                "The task runner must be started before submitting work."
            )

        # unpack the upstream call in order to cast Prefect futures to Dask futures
        # where possible to optimize Dask task scheduling
        parameters = self._optimize_futures(parameters)

        future = self._client.submit(
            task,
            parameters=parameters,
            wait_for=wait_for,
            dependencies=dependencies,
            return_type="state",
        )
        return PrefectDaskFuture(wrapped_future=future, task_run_id=future.task_run_id)

    @overload
    def map(
        self,
        task: "Task[P, Coroutine[Any, Any, R]]",
        parameters: Dict[str, Any],
        wait_for: Optional[Iterable[PrefectFuture]] = None,
    ) -> PrefectFutureList[PrefectDaskFuture[R]]:
        ...

    @overload
    def map(
        self,
        task: "Task[Any, R]",
        parameters: Dict[str, Any],
        wait_for: Optional[Iterable[PrefectFuture]] = None,
    ) -> PrefectFutureList[PrefectDaskFuture[R]]:
        ...

    def map(
        self,
        task: "Task",
        parameters: Dict[str, Any],
        wait_for: Optional[Iterable[PrefectFuture]] = None,
    ):
        return super().map(task, parameters, wait_for)

    def _optimize_futures(self, expr):
        def visit_fn(expr):
            if isinstance(expr, PrefectDaskFuture):
                dask_future = expr.wrapped_future
                if dask_future is not None:
                    return dask_future
            # Fallback to return the expression unaltered
            return expr

        return visit_collection(expr, visit_fn=visit_fn, return_data=True)

    def __enter__(self):
        """
        Start the task runner and prep for context exit.
        - Creates a cluster if an external address is not set.
        - Creates a client to connect to the cluster.
        """
        in_dask = False
        try:
            client = distributed.get_client()
            if client.cluster is not None:
                self._cluster = client.cluster
            elif client.scheduler is not None:
                self.address = client.scheduler.address
            else:
                raise RuntimeError("No global client found and no address provided")
            in_dask = True
        except ValueError:
            pass
        super().__enter__()
        exit_stack = self._exit_stack.__enter__()
        if self._cluster:
            self.logger.info(f"Connecting to existing Dask cluster {self._cluster}")
            self._connect_to = self._cluster
            if self.adapt_kwargs:
                self._cluster.adapt(**self.adapt_kwargs)
        elif self.address:
            self.logger.info(
                f"Connecting to an existing Dask cluster at {self.address}"
            )
            self._connect_to = self.address
        else:
            self.cluster_class = self.cluster_class or distributed.LocalCluster

            self.logger.info(
                f"Creating a new Dask cluster with "
                f"`{to_qualified_name(self.cluster_class)}`"
            )
            self._connect_to = self._cluster = exit_stack.enter_context(
                self.cluster_class(**self.cluster_kwargs)
            )

            if self.adapt_kwargs:
                maybe_coro = self._cluster.adapt(**self.adapt_kwargs)
                if inspect.isawaitable(maybe_coro):
                    run_coro_as_sync(maybe_coro)

        self._client = exit_stack.enter_context(
            PrefectDaskClient(self._connect_to, **self.client_kwargs)
        )

        if self._client.dashboard_link and not in_dask:
            self.logger.info(
                f"The Dask dashboard is available at {self._client.dashboard_link}",
            )

        return self

    def __exit__(self, *args):
        self._exit_stack.__exit__(*args)
        super().__exit__(*args)



import sys

import dask
import pytest
from distributed import Client
from prefect_dask import DaskTaskRunner, get_async_dask_client, get_dask_client

from prefect import flow, task

pytestmark = pytest.mark.skip(
    reason="Skipping tests while task runner is being rewritten"
)


class TestDaskSyncClient:
    def test_from_task(self):
        @task
        def test_task():
            delayed_num = dask.delayed(42)
            with get_dask_client() as client:
                assert isinstance(client, Client)
                result = client.compute(delayed_num).result()
            return result

        @flow(task_runner=DaskTaskRunner)
        def test_flow():
            future = test_task.submit()
            return future.result()

        assert test_flow() == 42

    def test_from_flow(self):
        @flow(task_runner=DaskTaskRunner)
        def test_flow():
            delayed_num = dask.delayed(42)
            with get_dask_client() as client:
                assert isinstance(client, Client)
                result = client.compute(delayed_num).result()
            return result

        assert test_flow() == 42

    def test_outside_run_context(self):
        delayed_num = dask.delayed(42)
        with get_dask_client() as client:
            assert isinstance(client, Client)
            result = client.compute(delayed_num).result()
        assert result == 42

    @pytest.mark.parametrize("timeout", [None, 8])
    def test_include_timeout(self, timeout):
        delayed_num = dask.delayed(42)
        with get_dask_client(timeout=timeout) as client:
            assert isinstance(client, Client)
            if timeout is not None:
                assert client._timeout == timeout
            result = client.compute(delayed_num).result()
        assert result == 42


class TestDaskAsyncClient:
    async def test_from_task(self):
        @task
        async def test_task():
            delayed_num = dask.delayed(42)
            async with get_async_dask_client() as client:
                assert isinstance(client, Client)
                result = await client.compute(delayed_num).result()
            return result

        @flow(task_runner=DaskTaskRunner)
        async def test_flow():
            future = test_task.submit()
            return future.result()

        assert (await test_flow()) == 42

    def test_from_sync_task_error(self):
        @task
        def test_task():
            with get_async_dask_client():
                pass

        @flow(task_runner=DaskTaskRunner)
        def test_flow():
            return test_task.submit()

        if sys.version_info < (3, 11):
            with pytest.raises(AttributeError, match="__enter__"):
                test_flow()
        else:
            with pytest.raises(
                TypeError, match="not support the context manager protocol"
            ):
                test_flow()

    async def test_from_flow(self):
        @flow(task_runner=DaskTaskRunner)
        async def test_flow():
            delayed_num = dask.delayed(42)
            async with get_async_dask_client() as client:
                assert isinstance(client, Client)
                result = await client.compute(delayed_num).result()
            return result

        assert (await test_flow()) == 42

    def test_from_sync_flow_error(self):
        @flow(task_runner=DaskTaskRunner)
        def test_flow():
            with get_async_dask_client():
                pass

        if sys.version_info < (3, 11):
            with pytest.raises(AttributeError, match="__enter__"):
                test_flow()
        else:
            with pytest.raises(
                TypeError, match="not support the context manager protocol"
            ):
                test_flow()

    async def test_outside_run_context(self):
        delayed_num = dask.delayed(42)
        async with get_async_dask_client() as client:
            assert isinstance(client, Client)
            result = await client.compute(delayed_num).result()
        assert result == 42

    @pytest.mark.parametrize("timeout", [None, 8])
    async def test_include_timeout(self, timeout):
        delayed_num = dask.delayed(42)
        async with get_async_dask_client(timeout=timeout) as client:
            assert isinstance(client, Client)
            if timeout is not None:
                assert client._timeout == timeout
            result = await client.compute(delayed_num).result()
        assert result == 42



import asyncio
import time
from typing import List

import distributed
import pytest
from distributed import LocalCluster
from prefect_dask import DaskTaskRunner
from prefect_dask.task_runners import PrefectDaskFuture

from prefect import flow, task
from prefect.futures import as_completed
from prefect.server.schemas.states import StateType
from prefect.states import State
from prefect.testing.fixtures import (  # noqa: F401
    hosted_api_server,
    use_hosted_api_server,
)


@pytest.fixture
def dask_task_runner_with_existing_cluster(use_hosted_api_server):  # noqa
    """
    Generate a dask task runner that's connected to a local cluster
    """
    with distributed.LocalCluster(n_workers=2) as cluster:
        yield DaskTaskRunner(cluster=cluster)


@pytest.fixture
def dask_task_runner_with_existing_cluster_address(use_hosted_api_server):  # noqa
    """
    Generate a dask task runner that's connected to a local cluster
    """
    with distributed.LocalCluster(n_workers=2) as cluster:
        with distributed.Client(cluster) as client:
            address = client.scheduler.address
            yield DaskTaskRunner(address=address)


@pytest.fixture
def dask_task_runner_with_process_pool(use_hosted_api_server):  # noqa
    yield DaskTaskRunner(cluster_kwargs={"processes": True})


@pytest.fixture
def dask_task_runner_with_thread_pool(use_hosted_api_server):  # noqa
    yield DaskTaskRunner(cluster_kwargs={"processes": False})


@pytest.fixture
def default_dask_task_runner(use_hosted_api_server):  # noqa
    yield DaskTaskRunner()


class TestDaskTaskRunner:
    @pytest.fixture(
        params=[
            default_dask_task_runner,
            # Existing cluster works outside of tests, but fails with serialization errors in tests
            # dask_task_runner_with_existing_cluster, :TODO Fix serialization errors in these tests
            dask_task_runner_with_existing_cluster_address,
            dask_task_runner_with_process_pool,
            dask_task_runner_with_thread_pool,
        ]
    )
    def task_runner(self, request):
        yield request.getfixturevalue(
            request.param._pytestfixturefunction.name or request.param.__name__
        )

    async def test_duplicate(self, task_runner):
        new = task_runner.duplicate()
        assert new == task_runner
        assert new is not task_runner

    async def test_successful_flow_run(self, task_runner):
        @task
        def task_a():
            return "a"

        @task
        def task_b():
            return "b"

        @task
        def task_c(b):
            return b + "c"

        @flow(version="test", task_runner=task_runner)
        def test_flow():
            a = task_a.submit()
            b = task_b.submit()
            c = task_c.submit(b)
            return a, b, c

        a, b, c = test_flow()
        assert await a.result() == "a"
        assert await b.result() == "b"
        assert await c.result() == "bc"

    async def test_failing_flow_run(self, task_runner):
        @task
        def task_a():
            raise RuntimeError("This task fails!")

        @task
        def task_b():
            raise ValueError("This task fails and passes data downstream!")

        @task
        def task_c(b):
            # This task attempts to use the upstream data and should fail too
            return b + "c"

        @flow(version="test", task_runner=task_runner)
        def test_flow():
            a = task_a.submit()
            b = task_b.submit()
            c = task_c.submit(b)
            d = task_c.submit(c)

            return a, b, c, d

        state = test_flow(return_state=True)

        assert state.is_failed()
        result = await state.result(raise_on_failure=False)
        a, b, c, d = result
        with pytest.raises(RuntimeError, match="This task fails!"):
            await a.result()
        with pytest.raises(
            ValueError, match="This task fails and passes data downstream"
        ):
            await b.result()

        assert c.is_pending()
        assert c.name == "NotReady"
        assert (
            f"Upstream task run '{b.state_details.task_run_id}' did not reach a"
            " 'COMPLETED' state" in c.message
        )

        assert d.is_pending()
        assert d.name == "NotReady"
        assert (
            f"Upstream task run '{c.state_details.task_run_id}' did not reach a"
            " 'COMPLETED' state" in d.message
        )

    async def test_async_tasks(self, task_runner):
        @task
        async def task_a():
            return "a"

        @task
        async def task_b():
            return "b"

        @task
        async def task_c(b):
            return b + "c"

        @flow(version="test", task_runner=task_runner)
        async def test_flow():
            a = task_a.submit()
            b = task_b.submit()
            c = task_c.submit(b)
            return a, b, c

        a, b, c = await test_flow()
        assert await a.result() == "a"
        assert await b.result() == "b"
        assert await c.result() == "bc"

    async def test_submit_and_wait(self, task_runner):
        @task
        async def task_a():
            return "a"

        async def fake_orchestrate_task_run(example_kwarg):
            return State(
                type=StateType.COMPLETED,
                data=example_kwarg,
            )

        with task_runner:
            future = task_runner.submit(task_a, parameters={}, wait_for=[])
            future.wait()
            state = future.state
            assert await state.result() == "a"

    async def test_async_task_timeout(self, task_runner):
        @task(timeout_seconds=0.1)
        async def my_timeout_task():
            await asyncio.sleep(2)
            return 42

        @task
        async def my_dependent_task(task_res):
            return 1764

        @task
        async def my_independent_task():
            return 74088

        @flow(version="test", task_runner=task_runner)
        async def test_flow():
            a = my_timeout_task.submit()
            b = my_dependent_task.submit(a)
            c = my_independent_task.submit()

            return a, b, c

        state = await test_flow(return_state=True)

        assert state.is_failed()
        ax, bx, cx = await state.result(raise_on_failure=False)
        assert ax.type == StateType.FAILED
        assert bx.type == StateType.PENDING
        assert cx.type == StateType.COMPLETED

    async def test_sync_task_timeout(self, task_runner):
        @task(timeout_seconds=1)
        def my_timeout_task():
            time.sleep(2)
            return 42

        @task
        def my_dependent_task(task_res):
            return 1764

        @task
        def my_independent_task():
            return 74088

        @flow(version="test", task_runner=task_runner)
        def test_flow():
            a = my_timeout_task.submit()
            b = my_dependent_task.submit(a)
            c = my_independent_task.submit()

            return a, b, c

        state = test_flow(return_state=True)

        assert state.is_failed()
        ax, bx, cx = await state.result(raise_on_failure=False)
        assert ax.type == StateType.FAILED
        assert bx.type == StateType.PENDING
        assert cx.type == StateType.COMPLETED

    def test_as_completed_yields_correct_order(self, task_runner):
        @task
        def sleep_task(seconds):
            time.sleep(seconds)
            return seconds

        timings = [1, 5, 10]
        with task_runner:
            done_futures = []
            futures = [
                task_runner.submit(
                    sleep_task, parameters={"seconds": seconds}, wait_for=[]
                )
                for seconds in reversed(timings)
            ]
            for future in as_completed(futures=futures):
                done_futures.append(future.result())
            assert done_futures == timings

    async def test_wait_captures_exceptions_as_crashed_state(self, task_runner):
        """
        Dask wraps the exception, interrupts will result in "Cancelled" tasks
        or "Killed" workers while normal errors will result in the raw error with Dask.
        We care more about the crash detection and
        lack of re-raise here than the equality of the exception.
        """
        if "processes" in task_runner.cluster_kwargs:
            pytest.skip("This will abort the run for a Dask cluster using processes.")

        @task
        def task_a():
            raise KeyboardInterrupt()

        with task_runner:
            future = task_runner.submit(
                task=task_a,
                parameters={},
                wait_for=[],
            )

            future.wait()
            assert future.state.type == StateType.CRASHED

    def test_dask_task_key_has_prefect_task_name(self):
        task_runner = DaskTaskRunner()

        @task
        def my_task():
            return 1

        futures = []

        @flow(task_runner=task_runner)
        def my_flow():
            futures.append(my_task.submit())
            futures.append(my_task.submit())
            futures.append(my_task.submit())

        my_flow()
        # ensure task run name is in key
        assert all(
            future.wrapped_future.key.startswith("my_task-") for future in futures
        )

    async def test_dask_cluster_adapt_is_properly_called(self):
        # mock of cluster instances with synchronous adapt method like
        # dask_kubernetes.classic.kubecluster.KubeCluster
        class MockDaskCluster(LocalCluster):
            def __init__(self, asynchronous: bool = False):
                self._adapt_called = False
                super().__init__(asynchronous=asynchronous)

            def adapt(self, **kwargs):
                self._adapt_called = True

        # mock of cluster instances with asynchronous adapt method like
        # dask_kubernetes.operator.kubecluster.KubeCluster
        class AsyncMockDaskCluster(MockDaskCluster):
            async def adapt(self, **kwargs):
                self._adapt_called = True

        for task_runner_class in [MockDaskCluster, AsyncMockDaskCluster]:
            # the adapt_kwargs argument triggers the calls to the adapt method
            task_runner = DaskTaskRunner(
                cluster_class=task_runner_class,
                adapt_kwargs={"minimum": 1, "maximum": 1},
            )
            with task_runner:
                assert task_runner._cluster._adapt_called

    def test_warns_if_future_garbage_collection_before_resolving(
        self, caplog, task_runner
    ):
        @task
        def test_task():
            return 42

        @flow(task_runner=task_runner)
        def test_flow():
            for _ in range(10):
                test_task.submit()

        test_flow()

        assert "A future was garbage collected before it resolved" in caplog.text

    def test_does_not_warn_if_future_resolved_when_garbage_collected(
        self, task_runner, caplog
    ):
        @task
        def test_task():
            return 42

        @flow(task_runner=task_runner)
        def test_flow():
            futures = [test_task.submit() for _ in range(10)]
            for future in futures:
                future.wait()

        test_flow()

        assert "A future was garbage collected before it resolved" not in caplog.text

    class TestInputArguments:
        async def test_dataclasses_can_be_passed_to_task_runners(self, task_runner):
            """
            this is a regression test for https://github.com/PrefectHQ/prefect/issues/6905
            """
            from dataclasses import dataclass

            @dataclass
            class Foo:
                value: int

            @task
            def get_dataclass_values(n: int):
                return [Foo(value=i) for i in range(n)]

            @task
            def print_foo(x: Foo) -> Foo:
                print(x)
                return x

            @flow(task_runner=task_runner)
            def test_dask_flow(n: int = 3) -> List[Foo]:
                foos = get_dataclass_values(n)
                future = print_foo.submit(foos[0])
                futures = print_foo.map(foos)

                return [fut.result() for fut in futures + [future]]

            results = test_dask_flow()

            assert results == [Foo(value=i) for i in range(3)] + [Foo(value=0)]

    def test_nested_task_submission(self, task_runner):
        @task
        def nested_task():
            return "nested task"

        @task
        def parent_task():
            nested_task_future = nested_task.submit()
            assert isinstance(nested_task_future, PrefectDaskFuture)
            return nested_task_future.result()

        @flow(task_runner=task_runner)
        def umbrella_flow():
            future = parent_task.submit()
            assert isinstance(future, PrefectDaskFuture)
            return future.result()

        assert umbrella_flow() == "nested task"



import pytest
from prefect_dask.client import PrefectDaskClient

from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import (
    FlowRunFilter,
    FlowRunFilterId,
    TaskRunFilter,
    TaskRunFilterId,
)
from prefect.client.schemas.objects import TaskRunResult
from prefect.client.schemas.sorting import TaskRunSort
from prefect.context import FlowRunContext
from prefect.flows import flow
from prefect.tasks import task
from prefect.testing.fixtures import (  # noqa: F401
    hosted_api_server,
    use_hosted_api_server,
)

pytestmark = pytest.mark.usefixtures("use_hosted_api_server")


class TestSubmit:
    async def test_with_task(self):
        flow_run_id = None

        @task
        def test_task():
            return 42

        @flow
        def test_flow():
            nonlocal flow_run_id
            flow_run_id = FlowRunContext.get().flow_run.id
            with PrefectDaskClient() as client:
                future = client.submit(test_task)
                return future.result()

        assert test_flow() == 42

        prefect_client = get_client()
        assert flow_run_id is not None
        task_runs = await prefect_client.read_task_runs(
            flow_run_filter=FlowRunFilter(id=FlowRunFilterId(any_=[flow_run_id]))
        )
        assert len(task_runs) == 1

    async def test_with_function(self):
        def func():
            return 42

        with PrefectDaskClient() as client:
            future = client.submit(func)
            assert future.result() == 42

    async def test_tracks_dependencies(self):
        flow_run_id = None

        @task
        def test_task(x):
            return x

        @flow
        def test_flow():
            nonlocal flow_run_id
            flow_run_id = FlowRunContext.get().flow_run.id
            with PrefectDaskClient() as client:
                future1 = client.submit(test_task, 42)
                future2 = client.submit(test_task, future1)
                return future2.result()

        assert test_flow() == 42

        prefect_client = get_client()
        assert flow_run_id is not None
        task_runs = await prefect_client.read_task_runs(
            flow_run_filter=FlowRunFilter(id=FlowRunFilterId(any_=[flow_run_id])),
            sort=TaskRunSort.END_TIME_DESC,
        )
        assert len(task_runs) == 2
        assert task_runs[0].task_inputs == {"x": [TaskRunResult(id=task_runs[1].id)]}


class TestMap:
    async def test_with_task(self):
        flow_run_id = None

        @task
        def test_task(x):
            return x

        @flow
        def test_flow():
            nonlocal flow_run_id
            flow_run_id = FlowRunContext.get().flow_run.id
            with PrefectDaskClient() as client:
                futures = client.map(test_task, [1, 2, 3])
                return [future.result() for future in futures]

        assert test_flow() == [1, 2, 3]

        prefect_client = get_client()
        assert flow_run_id is not None
        task_runs = await prefect_client.read_task_runs(
            flow_run_filter=FlowRunFilter(id=FlowRunFilterId(any_=[flow_run_id]))
        )
        assert len(task_runs) == 3

    async def test_with_function(self):
        def func(x):
            return x

        with PrefectDaskClient() as client:
            futures = client.map(func, [1, 2, 3])
            assert [future.result() for future in futures] == [1, 2, 3]

    async def test_tracks_dependencies(self):
        flow_run_id = None

        @task
        def test_task(x):
            return x

        @flow
        def test_flow():
            nonlocal flow_run_id
            flow_run_id = FlowRunContext.get().flow_run.id
            with PrefectDaskClient() as client:
                future1 = client.submit(test_task, 42)
                future2 = client.submit(test_task, 42)
                future3 = client.submit(test_task, 42)
                futures = client.map(test_task, [future1, future2, future3])
                return [future.result() for future in futures], [
                    future.task_run_id for future in futures
                ]

        result, task_run_ids = test_flow()
        assert result == [42, 42, 42]

        prefect_client = get_client()
        assert flow_run_id is not None
        task_runs = await prefect_client.read_task_runs(
            task_run_filter=TaskRunFilter(id=TaskRunFilterId(any_=task_run_ids)),
            sort=TaskRunSort.END_TIME_DESC,
        )
        for task_run in task_runs:
            # difficult to deterministically assert the order of task runs, so we'll check to make sure
            # each task run has a single input
            assert len(task_run.task_inputs["x"]) == 1



import pytest

from prefect.blocks.core import Block
from prefect.testing.standard_test_suites import BlockStandardTestSuite
from prefect.utilities.dispatch import get_registry_for_type
from prefect.utilities.importtools import to_qualified_name


def find_module_blocks():
    blocks = get_registry_for_type(Block)
    module_blocks = [
        block
        for block in blocks.values()
        if to_qualified_name(block).startswith("prefect_dask")
    ]
    return module_blocks


@pytest.mark.parametrize("block", find_module_blocks())
class TestAllBlocksAdhereToStandards(BlockStandardTestSuite):
    @pytest.fixture
    def block(self, block):
        return block



from . import _version
from .task_runners import DaskTaskRunner  # noqa
from .client import PrefectDaskClient  # noqa
from .utils import get_dask_client, get_async_dask_client  # noqa

__version__ = _version.__version__



import asyncio
from functools import wraps
from uuid import uuid4

from distributed import Client, Future

from prefect.context import serialize_context
from prefect.task_engine import run_task_async, run_task_sync
from prefect.tasks import Task
from prefect.utilities.callables import get_call_parameters
from prefect.utilities.engine import collect_task_run_inputs_sync


class PrefectDaskClient(Client):
    def submit(
        self,
        func,
        *args,
        key=None,
        workers=None,
        resources=None,
        retries=None,
        priority=0,
        fifo_timeout="100 ms",
        allow_other_workers=False,
        actor=False,
        actors=False,
        pure=True,
        **kwargs,
    ):
        if isinstance(func, Task):
            run_task_kwargs = {}
            run_task_kwargs["task"] = func
            run_task_kwargs["task_run_id"] = uuid4()
            run_task_kwargs["context"] = serialize_context()

            passed_dependencies = kwargs.pop("dependencies", None)
            run_task_kwargs["wait_for"] = kwargs.pop("wait_for", None)
            run_task_kwargs["return_type"] = kwargs.pop("return_type", "result")
            if (parameters := kwargs.get("parameters")) is None:
                # If parameters are not provided, we need to extract them from the function.
                # This case is when the PrefectDistributedClient is used directly without
                # the DaskTaskRunner.
                parameters = get_call_parameters(func, args, kwargs)
            run_task_kwargs["parameters"] = parameters
            dependencies = {
                k: collect_task_run_inputs_sync(v, future_cls=Future)
                for k, v in parameters.items()
            }
            if passed_dependencies:
                dependencies = {
                    k: v.union(passed_dependencies.get(k, set()))
                    for k, v in dependencies.items()
                }
            run_task_kwargs["dependencies"] = dependencies

            @wraps(func)
            def wrapper_func(*args, **kwargs):
                if func.isasync:
                    return asyncio.run(run_task_async(*args, **kwargs))
                else:
                    return run_task_sync(*args, **kwargs)

            future = super().submit(
                wrapper_func,
                key=key,
                workers=workers,
                resources=resources,
                retries=retries,
                priority=priority,
                fifo_timeout=fifo_timeout,
                allow_other_workers=allow_other_workers,
                actor=actor,
                actors=actors,
                pure=pure,
                **run_task_kwargs,
            )

            future.task_run_id = run_task_kwargs["task_run_id"]
            return future
        else:
            return super().submit(
                func,
                *args,
                key=key,
                workers=workers,
                resources=resources,
                retries=retries,
                priority=priority,
                fifo_timeout=fifo_timeout,
                allow_other_workers=allow_other_workers,
                actor=actor,
                actors=actors,
                pure=pure,
                **kwargs,
            )

    def map(
        self,
        func,
        *iterables,
        key=None,
        workers=None,
        retries=None,
        resources=None,
        priority=0,
        allow_other_workers=False,
        fifo_timeout="100 ms",
        actor=False,
        actors=False,
        pure=True,
        batch_size=None,
        **kwargs,
    ):
        if isinstance(func, Task):
            args_list = zip(*iterables)
            futures = []
            for args in args_list:
                futures.append(
                    self.submit(
                        func,
                        *args,
                        key=key,
                        workers=workers,
                        resources=resources,
                        retries=retries,
                        priority=priority,
                        fifo_timeout=fifo_timeout,
                        allow_other_workers=allow_other_workers,
                        actor=actor,
                        actors=actors,
                        pure=pure,
                        **kwargs,
                    )
                )
            return futures
        else:
            return super().map(
                func,
                *iterables,
                key=key,
                workers=workers,
                retries=retries,
                resources=resources,
                priority=priority,
                allow_other_workers=allow_other_workers,
                fifo_timeout=fifo_timeout,
                actor=actor,
                actors=actors,
                pure=pure,
                batch_size=batch_size,
                **kwargs,
            )
